{
 "cells": [
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "---\n",
    "title: \"📓 Week 10 - Student Notebook\"\n",
    "subtitle: Introduction to Natural Language Processing in Python\n",
    "author: \"MY NAME (MY CANDIDATE NUMBER)\"\n",
    "date: 24 March 2025\n",
    "---\n",
    "\n",
    "Let's begin by loading the relevant packages."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/html": [
       "\n",
       "            <div id=\"C4Cne6\"></div>\n",
       "            <script type=\"text/javascript\" data-lets-plot-script=\"library\">\n",
       "                if(!window.letsPlotCallQueue) {\n",
       "                    window.letsPlotCallQueue = [];\n",
       "                }; \n",
       "                window.letsPlotCall = function(f) {\n",
       "                    window.letsPlotCallQueue.push(f);\n",
       "                };\n",
       "                (function() {\n",
       "                    var script = document.createElement(\"script\");\n",
       "                    script.type = \"text/javascript\";\n",
       "                    script.src = \"https://cdn.jsdelivr.net/gh/JetBrains/lets-plot@v4.6.1/js-package/distr/lets-plot.min.js\";\n",
       "                    script.onload = function() {\n",
       "                        window.letsPlotCall = function(f) {f();};\n",
       "                        window.letsPlotCallQueue.forEach(function(f) {f();});\n",
       "                        window.letsPlotCallQueue = [];\n",
       "                        \n",
       "                    };\n",
       "                    script.onerror = function(event) {\n",
       "                        window.letsPlotCall = function(f) {};    // noop\n",
       "                        window.letsPlotCallQueue = [];\n",
       "                        var div = document.createElement(\"div\");\n",
       "                        div.style.color = 'darkred';\n",
       "                        div.textContent = 'Error loading Lets-Plot JS';\n",
       "                        document.getElementById(\"C4Cne6\").appendChild(div);\n",
       "                    };\n",
       "                    var e = document.getElementById(\"C4Cne6\");\n",
       "                    e.appendChild(script);\n",
       "                })()\n",
       "            </script>\n",
       "            "
      ]
     },
     "metadata": {},
     "output_type": "display_data"
    }
   ],
   "source": [
    "import numpy as np\n",
    "import pandas as pd\n",
    "import seaborn as sns\n",
    "import matplotlib.pyplot as plt\n",
    "import spacy\n",
    "import gensim\n",
    "import shap\n",
    "from gensim import corpora\n",
    "from gensim.models.ldamodel import LdaModel\n",
    "from gensim.models.coherencemodel import CoherenceModel\n",
    "from sklearn.feature_extraction.text import CountVectorizer\n",
    "from sklearn.linear_model import LogisticRegression\n",
    "from sklearn.ensemble import RandomForestClassifier\n",
    "from sklearn.metrics import confusion_matrix, f1_score\n",
    "from sklearn.model_selection import train_test_split\n",
    "from wordcloud import WordCloud\n",
    "from lets_plot import *\n",
    "LetsPlot.setup_html()\n",
    "\n",
    "# Load spaCy stopwords\n",
    "nlp = spacy.load(\"en_core_web_sm\")\n",
    "stopwords = nlp.Defaults.stop_words"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "# Introducing a new data set (10 minutes)\n",
    "\n",
    "We'll begin by loading data on 2,563 statements on economic matters by the European Central Bank (ECB), available [here](https://github.com/Moritz-Pfeifer/CentralBankRoBERTa/blob/main/Data/Labelled_data/Sentiment/ECB_prelabelled_sent.csv).  Statements by the ECB offer valuable insights into the economic policies and decisions that shape the financial landscape of the European Union. The ECB plays a critical role in regulating monetary policy, controlling inflation, managing interest rates, and ensuring economic stability across member states. Understanding how the ECB communicates its strategies and decisions helps students grasp broader economic concepts like central banking, economic growth, and market reactions. Additionally, ECB statements can reveal the political dynamics at play within the European Union, highlighting how economic policies influence social issues such as unemployment, inequality, and public welfare.\n",
    "\n",
    "There are two variables:\n",
    "\n",
    "-   `text`: Economic statements.\n",
    "-   `sentiment`: Whether the statement is positive (1) or negative (0)."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "# Load data\n",
    "\n",
    "\n",
    "# Print the data frame\n"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "👉 **NOTE:** As you can see, we have a series of statements that convey information about a specific economic topic and an attached sentiment.  However, the text data, as it stands, is not very useful but, with a few Python commands can be transformed into useful features."
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "# Creating our first document feature matrix (20 minutes)\n",
    "\n",
    "A document feature matrix (DFM) has an identical number of rows to the text column used to create it, and a given number of features, consisting of n-grams (n words seperated by white space) specified by the user.  The below code turns the text column into a document feature matrix.\n",
    "\n",
    "**🎯 Action Point:** Try your best to understand the code - don't worry if you don't, we will go over the code together."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "# Preprocess function: tokenize, remove punctuation, numbers, and stopwords\n",
    "def preprocess_text(text):\n",
    "    doc = nlp(text.lower())  # Lowercase text\n",
    "    tokens = [\n",
    "        token.lemma_ for token in doc \n",
    "        if not token.is_punct and not token.is_digit and not token.is_space \n",
    "        and token.text.lower() not in stopwords\n",
    "    ]\n",
    "    return \" \".join(tokens)  # Return cleaned text as a string\n",
    "\n",
    "# Apply preprocessing to each review\n",
    "\n",
    "\n",
    "# Create document-feature matrix (DFM) using the fit_transform method of CountVectorizer\n",
    "\n",
    "\n",
    "# Convert to a DataFrame for inspection\n",
    "\n"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "👉 **NOTE:** \n",
    "-   We create a count of each token in a given document, specifying that terms should occur in at least 10 documents in order to be included in the final DFM.\n",
    "-   We have used a pre-selected list of stopwords, however, you may need to employ a more nuanced list depending on your topic. For example, gender pronouns are included in the stopwords list which, in this case, could make sense but if you are studying topics relating to gender may not.\n",
    "\n",
    "## Some visualisations\n",
    "\n",
    "Here is a wordcloud, which visualises word frequencies by size in an aesthetically appealing format."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "# Create a corpus\n",
    "corpus = statements[\"text\"].to_list()\n",
    "\n",
    "# Collapse the corpus into a single list of words\n",
    "words = [word for text in corpus for word in text.split()]\n",
    "\n",
    "# Create value counts\n",
    "word_freq = pd.Series(words).value_counts()\n",
    "\n",
    "# Generate word cloud\n",
    "wordcloud = WordCloud(width=800, height=400, background_color='white').generate_from_frequencies(word_freq)\n",
    "\n",
    "# Display the word cloud\n",
    "plt.imshow(wordcloud, interpolation='bilinear')\n",
    "plt.axis(\"off\")\n",
    "plt.show()\n"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "🗣️ **CLASSROOM DISCUSSION:** What went wrong here? Edit the code to get more appropriate output."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "# Create a corpus\n",
    "corpus = statements[\"text_cleaned\"].to_list()\n",
    "\n",
    "# Collapse the corpus into a single list of words\n",
    "words = [word for text in corpus for word in text.split()]\n",
    "\n",
    "# Create value counts\n",
    "word_freq = pd.Series(words).value_counts()\n",
    "\n",
    "# Generate word cloud\n",
    "wordcloud = WordCloud(width=800, height=400, background_color='white').generate_from_frequencies(word_freq)\n",
    "\n",
    "# Display the word cloud\n",
    "plt.imshow(wordcloud, interpolation='bilinear')\n",
    "plt.axis(\"off\")\n",
    "plt.show()\n"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "To understand which words are most indicative of positive and negative sentiment, we can use `dfm` to predict our sentiment column using logistic regression.  We can then plot the top 10 largest absolute coefficients for the data."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "# Create our target using the sentiment column\n",
    "y = statements[\"sentiment\"]\n",
    "\n",
    "# Train a logistic regression model\n",
    "logit_model = LogisticRegression()\n",
    "logit_model.fit(dfm, y)\n",
    "\n",
    "# Get word importance (coefficients)\n",
    "feature_names = np.array(vectorizer.get_feature_names_out())\n",
    "coefficients = logit_model.coef_.flatten()\n",
    "\n",
    "# Create a data frame for plotting purposes\n",
    "coef_df = (pd.DataFrame(\n",
    "        {\n",
    "            \"feature\": feature_names,\n",
    "            \"coefficient\": coefficients,\n",
    "            \"abs_coefficient\": np.abs(coefficients),\n",
    "            \"sign\": np.where(coefficients > 0, \"positive\", \"negative\")\n",
    "        }\n",
    "    )\n",
    "    .groupby(\"sign\")\n",
    "    [[\"feature\",\"coefficient\",\"sign\",\"abs_coefficient\"]]\n",
    "    .apply(lambda x: x.nlargest(10, \"abs_coefficient\"))\n",
    "    .sort_values(\"coefficient\", ascending=True)\n",
    "    )\n",
    "\n",
    "# Plot results\n",
    "(\n",
    "    ggplot(coef_df, aes(\"coefficient\",\"feature\",fill=\"sign\")) +\n",
    "    geom_bar(stat=\"identity\") +\n",
    "    theme(legend_position=\"none\",\n",
    "          panel_grid_major_y = element_blank()) +\n",
    "    labs(x = \"Logit coefficient\", y = \"\")\n",
    ")"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "🗣️ **CLASSROOM DISCUSSION:** Which words are most indicative of positive / negative sentiment?"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "# A supervised learning application of NLP - predicting ECB sentiment (30 minutes)\n",
    "\n",
    "Let's print the first 5 rows of `dfm`."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "   ability  able  accelerate  access  accompany  account  accumulation  \\\n",
      "0        0     0           0       0          0        0             0   \n",
      "1        0     0           0       0          0        0             0   \n",
      "2        0     0           0       0          0        0             0   \n",
      "3        0     0           0       0          0        0             0   \n",
      "4        0     0           0       0          0        0             0   \n",
      "\n",
      "   achieve  act  action  ...  weaken  weakness  weigh  welfare  well  wide  \\\n",
      "0        0    0       0  ...       0         0      0        0     0     0   \n",
      "1        0    0       0  ...       0         0      0        0     0     0   \n",
      "2        0    0       0  ...       0         0      0        0     0     0   \n",
      "3        0    0       0  ...       0         0      0        0     0     0   \n",
      "4        0    0       0  ...       0         0      0        0     0     0   \n",
      "\n",
      "   work  world  worsen  year  \n",
      "0     0      0       0     0  \n",
      "1     0      0       0     0  \n",
      "2     0      0       0     0  \n",
      "3     0      0       0     0  \n",
      "4     0      0       0     0  \n",
      "\n",
      "[5 rows x 712 columns]\n"
     ]
    }
   ],
   "source": [
    "# Code here"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "We have 712 features that can be used for building a classification model. \n",
    "\n",
    "👉 **NOTE:** In doing this, we are using a Bag of Words (BoW) model - whereby a text (such as a sentence or document) is treated as an unordered collection (or \"bag\") of individual words, ignoring grammar and word order.  Each word in the text is represented by its frequency (or presence/absence) in the document. This means that the text is converted into a vector where each dimension corresponds to a unique word in the vocabulary (a list of all words across the dataset). The value in each dimension represents how often a word appears in the document.  The BoW model assumes that the context or syntax (word order, grammatical structure) does not matter. It only focuses on which words appear and how often.\n",
    "\n",
    "Next, let's create a train / test split, making sure to set a random seed and stratify the split by the outcome.  We will use `dfm` as comprising the totality of the features."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "# Code here"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "We can now specify a classification algorithm. We will opt for a Lasso model here, using an instantiation of `LogisticRegression`."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "# Instantiate a lasso classification model\n",
    "\n",
    "# Fit the model to the training data\n"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "Next, we can build a data frame of the top 10 absolute lasso coefficients and plot the results."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "# Create a data frame of the top 20 features\n",
    "\n",
    "\n",
    "# Plot the output\n"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "🗣️ **CLASSROOM DISCUSSION:**\n",
    "\n",
    "Which features are most important?\n",
    "\n",
    "We can also use Shapley values, both for each observation and in the aggregate.\n",
    "\n",
    "ℹ️ **Information:**\n",
    "\n",
    "Shapley values have a game theoretical underpinning. The basic idea is that we want to assign \"credit\" to a given feature for improving the performance of a model when compared to all possible sets (or \"coalitions\") of features. Instead of looking at the *global* contribution of each feature, as with variable importance, we compute the marginal contribution feature *j* makes to the prediction of an outcome for each observation, relative to the potential combination of features had the *j*th feature not been present. The Shapley value is then calculated as the average of all these marginal contributions across all observations. Shapley values can be calculated for all kinds of model, which makes them highly useful quantities of interest to calculate.\n",
    "\n",
    "For a more in-depth look at Shapley values, check out \"Interpretable Machine Learning\" from Christoph Molnar (click [here](https://christophm.github.io/interpretable-ml-book/shapley.html) for the relevant section).\n",
    "\n",
    "We can calculate Shapley values easily in Python."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "# Instantiate a Shapley values calculator, using the model and the training set features\n",
    "explainer = shap.Explainer(lasso_classifier, X_train)\n",
    "\n",
    "# Apply the calculator to the traing set features\n",
    "shap_values = explainer(X_train)\n",
    "\n",
    "# Create a \"bee swarm\" plot\n",
    "shap.plots.beeswarm(shap_values, max_display=15)"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "👉 **NOTE:**\n",
    "\n",
    "The \"beeswarm\" plot displays SHAP values per row of data.\n",
    "\n",
    "🗣️ **CLASSROOM DISCUSSION:**\n",
    "\n",
    "Which features are most important? Which make the most sense?\n",
    "\n",
    "Let's apply this model to the test set. Present a confusion matrix and evaluation metric of your choice."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "# Apply class predictions to the test set\n",
    "\n",
    "# Confusion matrix\n",
    "\n",
    "# Display confusion matrix\n",
    "\n",
    "# Print the evaluation metric of your choice "
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "# An unsupervised application of NLP - Finding ECB topics using Latent Dirichlet Allocation (30 minutes)\n",
    "\n",
    "Topic models are often used in social sciences to understand text data. These algorithms cluster words that are frequently used together across different documents into distinct topics (the number of which is set by the user). Each word is then ranked by how indicative they are of a given topic using some kind of score, of which there are many.\n",
    "\n",
    "We will pick 2 topics to begin with.  We will need to create 2 new data structures:\n",
    "\n",
    "-   A corpus dictionary, which assigns a unique token ID to all tokens across the corpus.\n",
    "-   A count of the number of times a token in the corpus dictionary appears in a given document.\n",
    "\n",
    "Let's create a corpus, which consists of a collection of documents."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "# Create a list of statements\n",
    "corpus = statements[\"text\"].to_list()\n",
    "\n",
    "# Print the first 10 entries in the corpus\n",
    "corpus[0:10]"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "We can then tokenise our corpus, meaning (in this case) that we can break our cleaned documents into distinct words that are not included in the stopwords list."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "# Create a tokenised corpus using a nested list comprehension\n",
    "corpus_tokenised = [\n",
    "    [word for word in document.split() if word not in stopwords]\n",
    "    for document in corpus\n",
    "]\n",
    "\n",
    "# Print the first tokenised corpus\n",
    "corpus_tokenised[0]"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "We can now create the corpus dictionary."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "# Create a corpus dictionary\n",
    "corpus_dictionary = corpora.Dictionary(corpus_tokenised)\n",
    "\n",
    "# Print the dictionary\n",
    "print(corpus_dictionary)"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "And the corpus count, which returns a list of tuples for each document, where the first element consists of the token ID and the second reports the count of that token in each document."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "# Create a statement bag of words (bow)\n",
    "corpus_count = [corpus_dictionary.doc2bow(text) for text in corpus_tokenised]\n",
    "\n",
    "# Print the count for the first two documents\n",
    "corpus_count[0:2]"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "We can now compile our LDA model."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "# Train an LDA model\n",
    "lda_model = LdaModel(corpus=corpus_count,\n",
    "                     id2word=corpus_dictionary,\n",
    "                     num_topics=2, \n",
    "                     random_state=123,\n",
    "                     passes=10,     \n",
    "                     iterations=50)"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "After having compiled our 2-topic model, we can now try to understand what each topic signifies \"under the hood\".  To do this, we can use a nested list comprehension where we employ the `get_topic_terms` method of `lda_model` to create a list of top 7 words for each topic."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "Topic 1: policy, euro, area, monetary, inflation, price, economic\n",
      "Topic 2: financial, market, bank, increase, risk, economy, banking\n"
     ]
    }
   ],
   "source": [
    "# List comprehension to isolate topics\n",
    "top_words = [\n",
    "    [corpus_dictionary[word_id] for word_id, _ in lda_model.get_topic_terms(topic_id, topn=7)]\n",
    "    for topic_id in range(lda_model.num_topics)\n",
    "]\n",
    "\n",
    "# Print results\n",
    "for i, words in enumerate(top_words):\n",
    "    print(f\"Topic {i+1}: {', '.join(words)}\")"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "🗣️ **CLASSROOM DISCUSSION:**\n",
    "\n",
    "Can we intuitively intepret the topics?\n",
    "\n",
    "👉 **NOTE:**\n",
    "\n",
    "-   The top 7 words are decided upon as the LDA model assigns a probability of each word appearing in a given topic.  The words that have the highest probabilities are therefore the most important.\n",
    "-   We use a trusty throwaway variable (`_`) in this context as `get_topic_terms` creates a tuple, with the word ID and word probability.  Since we do not want the word probability, we tell Python to ignore this.\n",
    "\n",
    "## How many topics?\n",
    "\n",
    "How did we arrive at only 2 topics?  Why not, say, 10? We can calculate performance metrics such as semantic coherence and perplexity:\n",
    "\n",
    "-   **Semantic coherence:** a measure of how commonly the most probable words in a topic co-occur. Note that one limitation of semantic coherence is that it is highest when the number of topics is low\n",
    "-   **Perplexity:** a measure of how well a model predicts a set of documents.\n",
    "\n",
    "We can set up a `for` loop over a range of topic numbers like so."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "# Instantiate empty lists for perplexity and coherence scores\n",
    "perplexities = []\n",
    "coherences = []\n",
    "\n",
    "# Create a range of topic numbers from 1 to 10\n",
    "topic_numbers = range(1,11)\n",
    "\n",
    "for topic in topic_numbers:\n",
    "    \n",
    "    # Train LDA model\n",
    "    lda_model = LdaModel(corpus=corpus_count,\n",
    "                         id2word=corpus_dictionary,\n",
    "                         num_topics=topic,  \n",
    "                         random_state=123,\n",
    "                         passes=10,     \n",
    "                         iterations=50)\n",
    "    \n",
    "    # Calculate perplexity\n",
    "    perplexity = lda_model.log_perplexity(corpus_count)\n",
    "    \n",
    "    # Calculate coherence\n",
    "    coherence = CoherenceModel(model=lda_model, corpus=corpus_count, coherence='u_mass').get_coherence()\n",
    "    \n",
    "    # Append values to list\n",
    "    perplexities.append(perplexity)\n",
    "    coherences.append(coherence)   "
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "After this, we can create a data frame of the results by passing a dictionary to `pd.DataFrame`.  All we then need to do is stack our scores on top of one another, for plotting purposes."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "# Create a data frame of results\n",
    "tm_df = pd.DataFrame(\n",
    "        {\n",
    "            \"topics\": topic_numbers,\n",
    "            \"Coherence\": coherences,\n",
    "            \"Perplexity\": perplexities\n",
    "        }\n",
    "    )\n",
    "\n",
    "# Pivot the data to a long data frame\n",
    "tm_melted_df = pd.melt(tm_df, id_vars=\"topics\", var_name = \"metric\", value_name= \"score\")\n",
    "\n",
    "# Plot the results\n",
    "(\n",
    "    ggplot(tm_melted_df, aes(\"topics\", \"score\")) +\n",
    "    geom_point() +\n",
    "    geom_line(linetype=\"dashed\") +\n",
    "    facet_wrap(facets=\"metric\", scales=\"free_y\") +\n",
    "    labs(x = \"Number of topics\", y = \"Score\",\n",
    "         title = \"Two topics is the most coherent and one of the least perplexing numbers!\",\n",
    "         subtitle = \"Experimenting with different topic numbers for LDA\\n\")\n",
    ")"
   ]
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Python 3",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.11.0"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 2
}
